///////////////////////////////////////////////////////////
//                    SvnInfoReader                      //
//                .svn      //
//                Copyright (C) 2007 Sagrer              //
//                             //
//                      LGPL v2.1                        //
//                                                       //
//                    0.1                    //
//                                                       //
//                  sagrer@yandex.ru                     //
///////////////////////////////////////////////////////////

//      , ....  :
// 1) Sagrer (sagrer@yandex.ru)

////////////////////////////////////////////////////////////////////////

unit SvnInfoReader;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, LCLTextFileInString, ExtraFunctionsLCL, ExtraFileUtilsLCL,
  DOM, XMLRead;
  
const
  SVNFormat_Max = 8;   //   SVN    .
  SVNFormat_1_3_ThisDir = 'svn:this_dir';   //  ThisDir   1-3 .
  SVNFormat_4_Newer_ThisDir = '';           //  ThisDir  4- .

type
  RSvnEntriesInfo = record         //      entries.
    ThisDirRevisionNumber : Integer;    //  .
  end;
  
  TSvnInfoReader = class
  private
  protected
    function Read_1_6_EntriesFormat(const FileName : AnsiString; const FormatNum : Integer) : Boolean;    // entries  1-6.
    function Read_7_8_EntriesFormat(const FileName : AnsiString; const FormatNum : Integer) : Boolean;    // entries  7-8.
  public
    //.
    ErrorMessage : AnsiString;     //   .
    EntriesInfo : RSvnEntriesInfo;

    //-...
    constructor Create; virtual;
    destructor Destroy; override;
    
    // 
    function IsSvnDir(const DirPath : AnsiString) : Boolean;               //     svn?
    function GetSvnDirFormat(const DirPath : AnsiString) : Integer;        //     .
    function GetSvnDirRevision(const DirPath : AnsiString) : Integer;      //    .
    procedure ResetEntriesInfo();                                          // EntriesInfo.
  end;

implementation

/////////////////////////////////////////////
//             TSvnInfoReaded              //
/////////////////////////////////////////////

//-----------------------------------------//
//        -...      //
//-----------------------------------------//

constructor TSvnInfoReader.Create;
begin
  //  ...

  //    ...
  ErrorMessage := '';

  //   ...
  
  //  .
  Self.ResetEntriesInfo();

end;

destructor TSvnInfoReader.Destroy;
begin
  // 
  Self.ResetEntriesInfo();

  //  
  inherited;
end;

//------------------------------------------//
//            ...           //
//------------------------------------------//

function TSvnInfoReader.Read_1_6_EntriesFormat(const FileName : AnsiString; const FormatNum : Integer) : Boolean;
// entries  1-6.
var
  EntriesFile : TXMLDocument;
  CurrNode : TDomNode;
  ThisDirStr : AnsiString;
  RevisionFinded : Boolean;
  
begin
  //.
  Result := true;
  RevisionFinded := false;
  if FormatNum <= 3 then begin         //    -    ThisDir
    ThisDirStr := SVNFormat_1_3_ThisDir;
  end
  else begin
    ThisDirStr := SVNFormat_4_Newer_ThisDir;
  end;
  
  //.      XML.    ?
  if FileExists(FileName) = true then begin
    // . .
    ReadXMLFile(EntriesFile, FileName);
    
    //  .
    CurrNode := EntriesFile.FirstChild;
    //  entry-.
    if CurrNode <> nil then CurrNode := CurrNode.FirstChild;
    //   -     ThisDir,  
    //  .
    repeat
      if CurrNode <> nil then begin
        //     entry.
        if CurrNode.Attributes.GetNamedItem('name').NodeValue = ThisDirStr then begin
          //,   . .
          Self.EntriesInfo.ThisDirRevisionNumber := StrToInt(CurrNode.Attributes.GetNamedItem('revision').NodeValue);
          RevisionFinded := true;
        end;
        //!
        CurrNode := CurrNode.NextSibling;
      end;
    until (CurrNode = nil) or (RevisionFinded = true);
    
    //   .
    EntriesFile.Free;
    
    // -   ?
    if RevisionFinded = false then begin
      //  .
      ErrorMessage := 'SVN info reader: Error! CurrentDirEntry with revision number was not found in -> '+FileName;
      Result := false;
    end;
  end
  else begin
    //,   .
    ErrorMessage := 'SVN info reader: Error! File not exists -> '+FileName;
    Result := false;
  end;
end;

function TSvnInfoReader.Read_7_8_EntriesFormat(const FileName : AnsiString; const FormatNum : Integer) : Boolean;
// entries  7-8.
var
  EntriesFile : TTextFileInString;
  ThisDirStr, RevStr : AnsiString;
  RevisionFinded : Boolean;
  
begin
  //.
  Result := true;
  RevisionFinded := false;
  ThisDirStr := SVNFormat_4_Newer_ThisDir;
  
  //.      .    ?
  if FileExists(FileName) = true then begin
    // . .
    EntriesFile := TTextFileInString.Create;
    if EntriesFile.Load(FileName) = true then begin
      //.    -    .
      EntriesFile.CurReadLn;
      //  ,    ThisDir     .
      repeat
        if (EntriesFile.FileString[EntriesFile.Cursor-2] = #12) or (EntriesFile.Cursor = 3) then begin
          //,   Entry.     .
          if EntriesFile.CurReadLn = ThisDirStr then begin
            //,   entry - ThisDir. ,      
            //   - 1 .
            EntriesFile.CurReadLn;
            //    .
            RevStr := EntriesFile.CurReadLn;
            if IsNumericStr(RevStr) = true then begin
              Self.EntriesInfo.ThisDirRevisionNumber := StrToInt(RevStr);
              RevisionFinded := true;
            end;
          end;
        end
        else begin
          EntriesFile.CurReadLn;
        end;
      until (EntriesFile.Eof = true) or (RevisionFinded = true);
    end;
    
    //   .
    EntriesFile.Free;

    // -   ?
    if RevisionFinded = false then begin
      //  .
      ErrorMessage := 'SVN info reader: Error! CurrentDirEntry with revision number was not found in -> '+FileName;
      Result := false;
    end;
  end
  else begin
    //,   .
    ErrorMessage := 'SVN info reader: Error! File not exists -> '+FileName;
    Result := false;
  end;
end;

//------------------------------------------//
//             ...            //
//------------------------------------------//

function TSvnInfoReader.IsSvnDir(const DirPath : AnsiString) : Boolean;
//     svn?
begin
  //.
  Result := false;
  
  if DirectoryExists(DirPath) = true then begin
    //   -      .svn
    if DirectoryExists(AddLastSlash(DirPath)+'.svn') = true then begin
      //        .
      if FileExists(AddLastSlash(DirPath)+'.svn\format') = true then begin
        //,   .
        Result := true;
      end;
    end;
  end;
  
end;

function TSvnInfoReader.GetSvnDirFormat(const DirPath : AnsiString) : Integer;
//     .
var
  FormatFile : TTextFileInString;
  VerString : AnsiString;
  
begin
  //.
  Result := 0;       // 0-   ,    .
  FormatFile := TTextFileInString.Create;
  
  // SVN-  ?
  if Self.IsSvnDir(DirPath) = true then begin
    //, .   .
    if FormatFile.Load(AddLastSlash(DirPath)+'.svn\format') = true then begin
      //.   .
      VerString := FormatFile.CurReadLn;
      if IsNumericStr(VerString) = true then begin
        //  .
        Result := StrToInt(VerString);
      end;
    end;
  end
  else begin
    ErrorMessage := 'SVN info reader: Error! Can''t find .svn directory in '+DirPath;
  end;
  
  // .
  FormatFile.Free;
end;

function TSvnInfoReader.GetSvnDirRevision(const DirPath : AnsiString) : Integer;
//    .
var
  SvnFormat : Integer;
  EntriesFileName : AnsiString;
  
begin
  //.
  Result := 0;      //   SVN    .
  
  //    .
  SvnFormat := Self.GetSvnDirFormat(DirPath);
  if (SvnFormat > 0) and (SvnFormat <= SVNFormat_Max) then begin
    //     .
    //  1-6 () -  entries   XML,   ,
    //   .   7  8 -    
    //,  7  8    entries  .
    // ,   -   2 .
    
    EntriesFileName := AddLastSlash(DirPath)+'.svn\entries';
    if FileExists(EntriesFileName) = true then begin
      //  .
      if SvnFormat <= 6 then begin
        if Self.Read_1_6_EntriesFormat(EntriesFileName, SvnFormat) = true then begin
          //.  .
          Result := Self.EntriesInfo.ThisDirRevisionNumber;
        end;
      end
      else begin
        if Self.Read_7_8_EntriesFormat(EntriesFileName, SvnFormat) = true then begin
          //.  .
          Result := Self.EntriesInfo.ThisDirRevisionNumber;
        end;
      end;
    end
    else begin
      ErrorMessage := 'SVN info reader: Error! File not exists -> '+EntriesFileName;
    end;
    
  end
  else begin
    ErrorMessage := 'SVN info reader: Error! Svn Format ('+IntToStr(SvnFormat)+') is too new - i don''t know how to read it!';
  end;
end;

procedure TSvnInfoReader.ResetEntriesInfo();
// EntriesInfo.
begin
  Self.EntriesInfo.ThisDirRevisionNumber := 0;    // ,     .
end;

end.

